//@line 4 "/builds/slave/m-b26_v1_2f-linux32_g-ntly-000/build/b2g/locales/en-US/b2g-l10n.js"

//@line 6 "/builds/slave/m-b26_v1_2f-linux32_g-ntly-000/build/b2g/locales/en-US/b2g-l10n.js"

pref("general.useragent.locale", "en-US");
