pref('browser.manifestURL', "app://system.gaiamobile.org/manifest.webapp");
pref('browser.homescreenURL', "app://system.gaiamobile.org/index.html");
pref('network.http.max-connections-per-server', 15);
pref('dom.mozInputMethod.enabled', true);
pref('ril.debugging.enabled', false);
pref('dom.mms.version', 17);
pref('network.dns.localDomains', "gaiamobile.org,calendar.gaiamobile.org,system.gaiamobile.org,email.gaiamobile.org,ringtones.gaiamobile.org,fm.gaiamobile.org,wappush.gaiamobile.org,costcontrol.gaiamobile.org,video.gaiamobile.org,wallpaper.gaiamobile.org,homescreen.gaiamobile.org,pdfjs.gaiamobile.org,clock.gaiamobile.org,camera.gaiamobile.org,gallery.gaiamobile.org,browser.gaiamobile.org,fl.gaiamobile.org,sms.gaiamobile.org,communications.gaiamobile.org,bluetooth.gaiamobile.org,music.gaiamobile.org,settings.gaiamobile.org,keyboard.gaiamobile.org,packstubtest.gaiamobile.org,mochitest.gaiamobile.org,hoststubtest.gaiamobile.org,marketplace.firefox.com.gaiamobile.org,twittershare.gaiamobile.org,ds-test.gaiamobile.org,geoloc.gaiamobile.org,membuster.gaiamobile.org,music2.gaiamobile.org,test-ime.gaiamobile.org,share-receiver.gaiamobile.org,template.gaiamobile.org,test-agent.gaiamobile.org,test-container.gaiamobile.org,test-keyboard-app.gaiamobile.org,test-otasp.gaiamobile.org,test-receiver-1.gaiamobile.org,test-receiver-2.gaiamobile.org,test-receiver-inline.gaiamobile.org,test-wappush.gaiamobile.org,uitest.gaiamobile.org");
pref("geo.gps.supl_server", "supl.izatcloud.net");
pref("geo.gps.supl_port", 22024);
pref("dom.payment.provider.0.name", "firefoxmarket");
pref("dom.payment.provider.0.description", "marketplace.firefox.com");
pref("dom.payment.provider.0.uri", "https://marketplace.firefox.com/mozpay/?req=");
pref("dom.payment.provider.0.type", "mozilla/payments/pay/v1");
pref("dom.payment.provider.0.requestMethod", "GET");

